/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2010 INRIA, USTL
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.control.property;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.api.control.NameController;

/**
 * Mixin layer for implementing the {@link SCAPropertyController} interface for
 * scaComposite components.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 1.1.2
 */
public abstract class SCACompositePropertyControllerMixin
implements SCAPropertyController {
    
    // -------------------------------------------------------------------------
    // Private constructor
    // -------------------------------------------------------------------------

    private SCACompositePropertyControllerMixin () {}

    // -------------------------------------------------------------------------
    // Implementation of the SCAPropertyController interface
    // -------------------------------------------------------------------------

    private Map<String,Object> values = new HashMap<String,Object>();
    private Map<String,Class<?>> types = new HashMap<String,Class<?>>();
    private Map<String,SCAPropertyController> promoters =
        new HashMap<String,SCAPropertyController>();
    
    /**
     * Set the type of the specified property. If the property has already been
     * set, the old value is lost, and the new value is recorded.
     * 
     * @param type   the property type
     * @param value  the property value
     */
    public void setType( String name, Class<?> type ) {
        
        SCAPropertyController promoter = promoters.get(name);
        if( promoter == null ) {
            types.put(name,type);
        }
        else {
            promoter.setType(name,type);
        }
    }
    
    /**
     * Set the value of the specified property. If the property has already been
     * set, the old value is lost, and the new value is recorded.
     * 
     * @param name   the property name
     * @param value  the property value
     */
    public void setValue( String name, Object value ) {
        
        SCAPropertyController promoter = promoters.get(name);
        if( promoter == null ) {
            values.put(name,value);
        }
        else {
            promoter.setValue(name,value);
        }
    }
    
    /**
     * Return the type of the specified property. Return <code>null</code> if
     * the property type has not been set.
     * 
     * @param name  the property name
     * @return      the property value
     */
    public Class<?> getType( String name ) {
        SCAPropertyController promoter = promoters.get(name);
        if( promoter == null ) {
            return types.get(name);
        }
        else {
            return promoter.getType(name);
        }
    }
    
    /**
     * Return the value of the specified property. Return <code>null</code> if
     * the property value has not been set.
     * 
     * @param name  the property name
     * @return      the property value
     */
    public Object getValue( String name ) {
        SCAPropertyController promoter = promoters.get(name);
        if( promoter == null ) {
            return values.get(name);
        }
        else {
            return promoter.getValue(name);
        }
    }
    
    /**
     * Return <code>true</code> if the specified property has been set.
     * 
     * @param name  the property name
     * @return      <code>true</code> if the property has been set,
     *              <code>false</code> otherwise
     */
    public boolean containsPropertyName( String name ) {
        SCAPropertyController promoter = promoters.get(name);
        if( promoter == null ) {
            return values.containsKey(name);
        }
        else {
            return promoter.containsPropertyName(name);
        }
    }

    /**
     * Return the names of the properties whose values have been set by invoking
     * {@link #setValue(String, Object)}.
     */
    public String[] getPropertyNames() {        
        Set<String> keys = values.keySet();
        String[] names = keys.toArray( new String[keys.size()] );
        return names;
    }
    
    /**
     * Return <code>true</code> if the specified property can be injected in the
     * content class.
     * 
     * @param name  the property name
     * @return      <code>true</code> if the property can be injected,
     *              <code>false</code> otherwise
     * @since 1.1.2
     */
    public boolean containsDeclaredPropertyName( String name ) {

        /*
         * For scaComposite components, we can only rely on properties declared
         * by scaPrimitives which would promote their properties.
         */
        Collection<SCAPropertyController> scapcs = promoters.values();
        for (SCAPropertyController scapc : scapcs) {
            if( scapc.containsDeclaredPropertyName(name) ) {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Return the names of the properties which can be injected in the content
     * class.
     */
    public String[] getDeclaredPropertyNames() {
        
        /*
         * For scaComposite components, we can only rely on properties declared
         * by scaPrimitives which would promote their properties.
         */
        Set<String> propnames = new HashSet<String>();
        Collection<SCAPropertyController> scapcs = promoters.values();
        for (SCAPropertyController scapc : scapcs) {
            String[] s = scapc.getDeclaredPropertyNames();
            propnames.addAll( Arrays.asList(s) );
        }
        
        String[] names = propnames.toArray( new String[propnames.size()] );
        return names;
    }
    
    /**
     * Return the type of the specified property, provided that this property
     * can be injected in the content class.
     * 
     * @param name  the property name
     * @return      the property type
     * @since 1.1.2
     */
    public Class<?> getDeclaredPropertyType( String name ) {

    	/*
         * For scaComposite components, we can only rely on properties declared
         * by scaPrimitives which would promote their properties.
         */
        Collection<SCAPropertyController> scapcs = promoters.values();
        for (SCAPropertyController scapc : scapcs) {
            if( scapc.containsDeclaredPropertyName(name) ) {
            	Class<?> type = scapc.getDeclaredPropertyType(name);
                return type;
            }
        }
        
    	return null;
    }
    
    /**
     * Set the reference of the property controller which promotes the specified
     * property to the current property controller.
     * 
     * @param name      the promoter property name
     * @param promoter  the promoter component or
     *                  <code>null</code> to unregister the promoter
     * @throws IllegalPromoterException
     *      thrown when attempting to set a cycle between property promoters
     */
    public void setPromoter( String name, SCAPropertyController promoter )
    throws IllegalPromoterException {
        
        promoters.put(name,promoter);

        if( promoter != null ) {
            SCAPropertyController peer = promoter.getPromoter(name);
            if( peer == this ) {
                String compname = _this_weaveableOptNC.getFcName();
                throw new IllegalPromoterException(compname);
            }
        }
    }
    
    /**
     * Return the reference of the property controller which promotes the
     * specified property. Return <code>null</code> if the property is managed
     * locally by the current property controller.
     * 
     * @param name  the promoter property name
     * @return      the promoter component or <code>null</code>
     */
    public SCAPropertyController getPromoter( String name ) {
        return promoters.get(name);
    }

    
    // -------------------------------------------------------------------------
    // Fields and methods required by the mixin class in the base class
    // -------------------------------------------------------------------------

    /**
     * The {@link NameController} interface of the component to which this
     * controller object belongs.
     */
    public NameController _this_weaveableOptNC;
}
